/*
 
 aHydro1p1b.cpp
 
 Copyright (c) Mohammad Nopoush and Michael Strickland
 
 GNU General Public License (GPLv3)
 See detailed text in license directory
 
 */

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <cstdio>
#include <cmath>

using namespace std;

#include "aHydro1p1b.h"
#include "equations.h"
#include "rfuncs.h"
#include "outputroutines.h"
#include "initialconditions.h"
#include "paramreader.h"
#include "memory.h"
#include "eos.h"
#include "freezeout.h"

using namespace std;

// these global vars are initialized from parameters file
// defaults set here are overridden by that file
int    NUM=20,STEPS=10000,UPDATE=10,SNAPUPDATE=100,INITCONDTYPE=0,STEPPER=1,IDEAL=0,USELAX=1,EOS=0,NA=208,NB=208,COMPUTEFREEZEOUT=0,ISOFREEZE=0;
double  A=0.05,EPS=0.001,TAU0=0.2,ETABAR=0.1,L0=1.0,ALPHAZ0=1+SMALL,ALPHAX0=1+SMALL,LAXFRAC=0.05,SIGMANN=62.0,Q=1.,TFO=0.761;
char EOSDIR[100],PARTICLEFILE[100];

// these hold the current values of the variables
double *thetap,*lambda,*alphax,*alphaz;

// these hold the updated values of the variables
double *Thetap,*Lambda,*Alphax,*Alphaz;

// these hold the RHS values needed for rk4 algorithm
double *rhs_thetap,*rhs_lambda,*rhs_alphax,*rhs_alphaz;

// these hold temporary values needed for rk4 algorithm
double *RK4_thetap,*RK4_lambda,*RK4_alphax,*RK4_alphaz;

// output files
fstream monitor_out;

// "copies" updated arrays using pointer swap
void copyDown() {
    
    double *tmp;
    
    tmp = lambda;
    lambda = Lambda;
    Lambda = tmp;
    
    tmp = alphax;
    alphax = Alphax;
    Alphax = tmp;
    
    tmp = alphaz;
    alphaz = Alphaz;
    Alphaz = tmp;
    
    tmp = thetap;
    thetap = Thetap;
    Thetap  = tmp;
    
}

// returns first dervative with respect to r of var
inline double ddr(double* var, int s) {
#if DERIVORDER == 2
    if (s==0) return (var[1]-var[0])/A; // right difference (first order)
    if (s==NUM-1) return (var[NUM-1]-var[NUM-2])/A; // left difference (first order)
    return (var[s+1] - var[s-1])/2./A; // fourth order
#elif DERIVORDER == 4
    if (s==0) return (var[1]-var[0])/A; // right difference (first order)
    if (s==1) return (var[s+1] - var[s-1])/2./A; // centered difference (second order)
    if (s==NUM-2) return (var[s+1] - var[s-1])/2./A; // centered difference (second order)
    if (s==NUM-1) return (var[NUM-1]-var[NUM-2])/A; // left difference (first order)
    return (8*(var[s+1] - var[s-1]) - (var[s+2] - var[s-2]))/12./A; // fourth order
#endif
}

inline double r(int s) {
    return (s - (NUM-1)/2.)*A;
}

void loadRHS(double t, double* llambda, double* lalphax, double* lalphaz, double* lthetap) {
    
    double drlambda,dralphax,dralphaz,drthetap;
    
    for (int s=0;s<NUM;s++) {
        
        //cout << "s = " << s << endl;
        
        drlambda = ddr(llambda,s);
        dralphax = ddr(lalphax,s);
        dralphaz = ddr(lalphaz,s);
        drthetap = ddr(lthetap,s);
        
        double lr = r(s); // compute value of r corresponding to s
        
        struct stateVars myStateVars = {t,lr,llambda[s],lalphax[s],lalphaz[s],lthetap[s],drlambda,dralphax,dralphaz,drthetap};
        
        struct stateFncs myStateFncs;
        
        loadStateFncs(&myStateVars,&myStateFncs);
        
        rhs_lambda[s] = lambdaRHS(&myStateVars,&myStateFncs);
        rhs_alphax[s] = alphaxRHS(&myStateVars,&myStateFncs);
        rhs_alphaz[s] = alphazRHS(&myStateVars,&myStateFncs);
        rhs_thetap[s] = thetapRHS(&myStateVars,&myStateFncs);
        
    }
}

void rkStep(double t, double* llambda, double* lalphax, double* lalphaz, double* lthetap, int c1, int c2, short int load) {
    loadRHS(t,llambda,lalphax,lalphaz,lthetap);
    for (int s=0;s<NUM;s++) {
        // lambda
        Lambda[s] += EPS*rhs_lambda[s]/c1;
        if (load) RK4_lambda[s] = lambda[s] + EPS*rhs_lambda[s]/c2;
        // alphax
        Alphax[s] += EPS*rhs_alphax[s]/c1;
        if (load) RK4_alphax[s] = alphax[s] + EPS*rhs_alphax[s]/c2;
        // alphaz
        Alphaz[s] += EPS*rhs_alphaz[s]/c1;
        if (load) RK4_alphaz[s] = alphaz[s] + EPS*rhs_alphaz[s]/c2;
        // thetap
        Thetap[s] += EPS*rhs_thetap[s]/c1;
        if (load) RK4_thetap[s] = thetap[s] + EPS*rhs_thetap[s]/c2;
    }
}

void laxSmooth(double *in, double *out) {
    int sm,sp;
    for (int s=0;s<NUM;s++) {
        sm = (s!=0) ? s-1 : s;
        sp = (s<NUM-1) ? s+1 : s;
        out[s] = LAXFRAC*0.5*(in[sm]+in[sp]) + (1-LAXFRAC)*in[s];
    }
}

// does on incremental step forward in time
void doInc(double t) {
    // set updated vars to current values
    for (int s=0;s<NUM;s++) {
        Lambda[s] = lambda[s];
        Alphax[s] = alphax[s];
        Alphaz[s] = alphaz[s];
        Thetap[s] = thetap[s];
    }
    // use specified algorithm for temporal update
    switch (STEPPER) {
        case 0:  // Euler Method
            rkStep(t,lambda,alphax,alphaz,thetap,1,1,0);
            break;
        case 1:  // 4th order Runge-Kutta
            // RK4 step 1
            rkStep(t,lambda,alphax,alphaz,thetap,6,2,1);
            // RK4 step 2
            rkStep(t+EPS/2,RK4_lambda,RK4_alphax,RK4_alphaz,RK4_thetap,3,2,1);
            // RK4 step 3
            rkStep(t+EPS/2,RK4_lambda,RK4_alphax,RK4_alphaz,RK4_thetap,3,1,1);
            // RK4 step 4
            rkStep(t+EPS,RK4_lambda,RK4_alphax,RK4_alphaz,RK4_thetap,6,1,0);
            break;
        default:
            cout << "Update algorithm not implemented.  Exiting." << endl;
            exit(-1);
    }
}

void Evolve() {
    double t = TAU0;
    for (int i=1;i<=STEPS+1;i++) {
        
        if (COMPUTEFREEZEOUT==1) {
            pushFreezeoutData(t);
            // check to see if the center has fallen below the freezeout temperature
            // and, if so, stop calculating and dump some info and snapshots
            if (Teff(alphax[NUM/2],alphaz[NUM/2],lambda[NUM/2]) < 0.5*TFO) {
                outputMeasurements(t);
                outputSnapshot(lambda,t,1,"lambda");
                outputSnapshot(alphax,t,1,"alphax");
                outputSnapshot(alphaz,t,1,"alphaz");
                outputSnapshot(thetap,t,1,"thetap");
                break;
            }
        }
        // output measurements and snapshots if appropriate
        if ((i-1)%UPDATE==0) {
            outputMeasurements(t);
        }
        if ((i-1)%SNAPUPDATE==0) {
            outputSnapshot(lambda,t,1,"lambda");
            outputSnapshot(alphax,t,1,"alphax");
            outputSnapshot(alphaz,t,1,"alphaz");
            outputSnapshot(thetap,t,1,"thetap");
            outputTeff(t);
        }
        // evolve variables eps forward in time storing updated fields in captial vars
        doInc(t);
        // increment time
        t = TAU0 + i*EPS;
        // copy fields from capital vars (updated) down to lowercase vars (current)
        if (USELAX) {
            laxSmooth(Lambda,lambda);
            laxSmooth(Thetap,thetap);
            laxSmooth(Alphax,alphax);
            laxSmooth(Alphaz,alphaz);
        } else {
            copyDown();
        }
                
    }
}

int main(int argc, char** argv) {
    
    print_line();
    
    // read parameters from file and command line
    readParametersFromFile("input/params.txt",1);
    if (argc>1) {
        print_line();
        cout << "Parameters from commandline" << endl;
        print_line();
        readParametersFromCommandLine(argc,argv,1);
    }
    // perform any processing of parameters necessary
    processParameters();
    
    // open files to put the data in first
    monitor_out.open("output/monitor.dat", ios::out);
    
    print_line();
    
    allocateMemory();
    
    initializeEoS();
    
    setInitialConditions();
    
    print_line();
    
    cout << "Spatial Step Size (A): " << A << " fm" << endl;
    
    cout << "Temporal Step Size [fm/c] (EPS): " << EPS << " fm/c" << endl;
    
    cout << "Box Size [fm] (A*" << NUM << "): " << A*NUM << " fm" << endl;
    
    print_line();
    
    cout.width(dwidth); cout << "tau [fm/c]";
    cout.width(dwidth); cout << "lambda [GeV]";
    cout.width(dwidth); cout << "alphax";
    cout.width(dwidth); cout << "alphaz";
    cout.width(dwidth); cout << "Teff [GeV]";
    cout << endl;
    
    print_line();
    
    Evolve();
    
    if (COMPUTEFREEZEOUT==1) doFreezeOut();
    
    print_line();
    cout << "Done.\n";
    print_line();
    
    // finally close files
    monitor_out.close();
    
    // free memory
    finalizeEoS();
    freeMemory();
    
    return 0;
}

