/*
 
 freezeout.cpp
 
 Copyright (c) Mohammad Nopoush and Michael Strickland
 
 GNU General Public License (GPLv3)
 See detailed text in license directory
 
 */

#include <fstream>
#include <iostream>
#include <sstream>
#include <cmath>
#include <vector>
#include <algorithm>
#include <gsl/gsl_spline.h>

using namespace std;

#include "aHydro1p1b.h"
#include "eos.h"
#include "freezeout.h"
#include "outputroutines.h"

vector<double> tauVec(0);
vector<double> rfoVec(0);
vector<double> lambdaVec(0);
vector<double> alphaxVec(0);
vector<double> alphazVec(0);
vector<double> thetapVec(0);

double zeta[ZETAPTS]; // "polar" angle in (tau-tau0)-r plane
double d[ZETAPTS]; // distance measure
double dp[ZETAPTS]; // derivative of distance measure
double L[ZETAPTS];
double AX[ZETAPTS];
double AZ[ZETAPTS];
double TP[ZETAPTS];

bool zeroFlag = false;

// variables for the interpolations

gsl_interp_accel *accD;
gsl_spline *splineD;

gsl_interp_accel *accLambda;
gsl_spline *splineLambda;

gsl_interp_accel *accAlphax;
gsl_spline *splineAlphax;

gsl_interp_accel *accAlphaz;
gsl_spline *splineAlphaz;

gsl_interp_accel *accThetap;
gsl_spline *splineThetap;

//////////////////////////////////////////////////////////////////////////////////////
// data types
//////////////////////////////////////////////////////////////////////////////////////

struct particleData {
    double mt, pt, y, phi, statistics;
};

struct fluidVars {
    double lambda, ax, az, thetap;
};

struct hypersurfaceVars {
    double d, dp, zeta, phi, theta;
};

//////////////////////////////////////////////////////////////////////////////////////
// support functions
//////////////////////////////////////////////////////////////////////////////////////

inline double r(int s) {
    return (s - (NUM-1)/2.)*A;
}

inline double ctrVal(double *array) {
    int h = NUM/2;
    return (array[h]+array[h-1])/2.;
}

inline double pdotu(particleData pd, fluidVars fv, hypersurfaceVars hs) {
    double varsigma = hs.d/tan(hs.theta)/SCALEFACTOR;
    return pd.mt*cosh(pd.y-varsigma)*cosh(fv.thetap) - pd.pt*cos(hs.phi-pd.phi)*sinh(fv.thetap);
}

inline double pdotx(particleData pd, fluidVars fv, hypersurfaceVars hs) {
    double varsigma = hs.d/tan(hs.theta)/SCALEFACTOR;
    return pd.mt*cosh(pd.y-varsigma)*sinh(fv.thetap) - pd.pt*cos(hs.phi-pd.phi)*cosh(fv.thetap);
}

inline double pdoty(particleData pd, fluidVars fv, hypersurfaceVars hs) {
    return pd.pt*sin(hs.phi-pd.phi);
}

inline double pdotz(particleData pd, fluidVars fv, hypersurfaceVars hs) {
    double varsigma = hs.d/tan(hs.theta)/SCALEFACTOR;
    return -pd.mt*sinh(pd.y-varsigma);
}

inline double pdSigma(particleData pd, fluidVars fv, hypersurfaceVars hs) {
    double varsigma = hs.d/tan(hs.theta)/SCALEFACTOR;
    double tau = TAU0 + hs.d*sin(hs.zeta);
    double temp  = hs.dp*(pd.pt*sin(hs.zeta)*cos(hs.phi-pd.phi)-pd.mt*cos(hs.zeta)*cosh(pd.y-varsigma));
    temp += hs.d*(pd.mt*sin(hs.zeta)*cosh(pd.y-varsigma) + pd.pt*cos(hs.zeta)*cos(hs.phi-pd.phi));
    return temp*tau*hs.d*hs.d*cos(hs.zeta)/sin(hs.theta)/sin(hs.theta)/SCALEFACTOR;
}

inline double udSigma(fluidVars fv, hypersurfaceVars hs) {
    double varsigma = hs.d/tan(hs.theta)/SCALEFACTOR;
    double tau = TAU0 + hs.d*sin(hs.zeta);
    double temp  = hs.d*(cosh(fv.thetap)*sin(hs.zeta) + cos(hs.zeta)*sinh(fv.thetap));
    temp += hs.dp*(sin(hs.zeta)*sinh(fv.thetap)-cos(hs.zeta)*cosh(fv.thetap));
    return temp*tau*hs.d*hs.d*cos(hs.zeta)/sin(hs.theta)/sin(hs.theta)/SCALEFACTOR;
}

inline double distFnc(particleData pd, fluidVars fv, hypersurfaceVars hs) {
    if (ISOFREEZE==1) {
        double T = Teff(fv.ax,fv.az,fv.lambda);
        double pu = pdotu(pd,fv,hs);
        return 1/(exp(pu/T)+pd.statistics);
    }
    double ay = fv.ax*fv.az/sqrt(3*pow(fv.ax*fv.az,2) - fv.az*fv.az - fv.ax*fv.ax);
    double xix = 1/fv.ax/fv.ax-1;
    double xiy = 1/ay/ay-1;
    double xiz = 1/fv.az/fv.az-1;
    double pu = pdotu(pd,fv,hs);
    double px = pdotx(pd,fv,hs);
    double py = pdoty(pd,fv,hs);
    double pz = pdotz(pd,fv,hs);
    return 1/(exp(sqrt(pu*pu + xix*px*px + xiy*py*py + xiz*pz*pz)/fv.lambda)+pd.statistics);
}

inline double nDens(fluidVars fv) {
    double ay = fv.ax*fv.az/sqrt(3*pow(fv.ax*fv.az,2) - fv.az*fv.az - fv.ax*fv.ax);
    return fv.ax*ay*fv.az*fv.lambda*fv.lambda*fv.lambda/M_PI/M_PI;
}

inline double freezeoutIntegrand(particleData pd, fluidVars fv, hypersurfaceVars hs) {
    return pdSigma(pd,fv,hs)*distFnc(pd,fv,hs);
}

inline double nDensIntegrand(fluidVars fv, hypersurfaceVars hs) {
    return udSigma(fv,hs)*nDens(fv);
}

//////////////////////////////////////////////////////////////////////////////////////

inline void pushPoint(double t, int s, double x) {
    tauVec.push_back(t-TAU0);
    rfoVec.push_back(r(s-1) + x*A);
    lambdaVec.push_back((1-x)*lambda[s-1] + x*lambda[s]);
    alphaxVec.push_back((1-x)*alphax[s-1] + x*alphax[s]);
    alphazVec.push_back((1-x)*alphaz[s-1] + x*alphaz[s]);
    thetapVec.push_back((1-x)*thetap[s-1] + x*thetap[s]);
}

void pushFreezeoutData(double t) {
    
    int s;
    double x;
    
    // if the center is below TFO then handle this carefully
    if (Teff(alphax[NUM/2],alphaz[NUM/2],lambda[NUM/2]) < TFO) {
        if (!zeroFlag) {
            tauVec.push_back(t-TAU0);
            rfoVec.push_back(0.);
            lambdaVec.push_back(ctrVal(lambda));
            alphaxVec.push_back(ctrVal(alphax));
            alphazVec.push_back(ctrVal(alphaz));
            thetapVec.push_back(ctrVal(thetap));
            zeroFlag=true;
        }
        // check for interior surface
        double tm,tp;
        tm = Teff(alphax[NUM/2],alphaz[NUM/2],lambda[NUM/2]);
        for (s=NUM/2; s<NUM; s++) {
            tp = Teff(alphax[s],alphaz[s],lambda[s]);
            if (tp > TFO) break;
            else tm = tp;
        }
        // check for exterior surface
        if (s!=NUM) {
            x = (tp != tm ? (TFO-tm)/(tp-tm) : 0.5); // fractional distance between s-1 and s
            //cout << "found interior surface 1 s=" << s << ", x=" << x << endl;
            pushPoint(t,s,x);
            // try coming in from the right edge to identify the exterior surface associated with the interior surface
            tp = Teff(alphax[NUM-1],alphaz[NUM-1],lambda[NUM-1]);
            for (s=NUM-1; s>=NUM/2; s--) {
                tm = Teff(alphax[s],alphaz[s],lambda[s]);
                if (tm > TFO) break;
                else tp = tm;
            }
            if (s!=NUM/2) {
                x = (tp != tm ? (TFO-tm)/(tp-tm) : 0.5); // fractional distance between s-1 and s
                //cout << "found interior surface 2 s=" << s << ", x=" << x << endl;
                pushPoint(t,s+1,x);
            }
        }
        return;
    }
    else {
        // check for exterior surface
        double tm,tp;
        tm = Teff(alphax[NUM/2-1],alphaz[NUM/2-1],lambda[NUM/2-1]);
        for (s=NUM/2; s<NUM; s++) {
            tp = Teff(alphax[s],alphaz[s],lambda[s]);
            if (tp < TFO) break;
            else tm = tp;
        }
        double x = (tp != tm ? (TFO-tm)/(tp-tm) : 0.5); // fractional distance between s-1 and s
        pushPoint(t,s,x);
        return;
    }
}

double deriv(double *x, double* y, int i) {
    if (i==0) return (y[1]-y[0])/(x[1]-x[0]);
    if (i==ZETAPTS-1) return (y[ZETAPTS-1]-y[ZETAPTS-2])/(x[ZETAPTS-1]-x[ZETAPTS-2]);
    return (y[i+1]-y[i-1])/(x[i+1]-x[i-1]);
}

struct sortArray {
    double val;
    int idx;
};

bool comparator(sortArray l, sortArray r)
{
    return l.val < r.val;
}

// loads freeze variable arrays
void loadFreezeoutVariables() {
    
    int nptsRaw = tauVec.size();
    double *zetaRaw = new double[nptsRaw];
    double *dRaw = new double[nptsRaw];
    
    for (int i=0;i<nptsRaw;i++) {
        double t = tauVec[i];
        double r = rfoVec[i];
        zetaRaw[i] = (r!=0 ? atan(t/r) : M_PI/2);
        dRaw[i] = sqrt(t*t+r*r);
    }
    
    // sort lists by zeta
    double *zetaSorted = new double[nptsRaw];
    double *dSorted = new double[nptsRaw];
    double *lambdaSorted = new double[nptsRaw];
    double *alphaxSorted = new double[nptsRaw];
    double *alphazSorted = new double[nptsRaw];
    double *thetapSorted = new double[nptsRaw];
    vector<sortArray> zetaSort;
    for (int i=0;i<nptsRaw; i++) {
        sortArray sa = {zetaRaw[i],i};
        zetaSort.push_back(sa);
    }
    std::sort(zetaSort.begin(), zetaSort.end(), comparator);
    for (int i=0;i<nptsRaw; i++) {
        sortArray sa = zetaSort[i];
        zetaSorted[i] = sa.val;
        dSorted[i] = dRaw[sa.idx];
        lambdaSorted[i] = lambdaVec[sa.idx];
        alphaxSorted[i] = alphaxVec[sa.idx];
        alphazSorted[i] = alphazVec[sa.idx];
        thetapSorted[i] = thetapVec[sa.idx];
    }
    
    // construct interpolating functions
    accD = gsl_interp_accel_alloc();
    splineD = gsl_spline_alloc(gsl_interp_cspline, nptsRaw);
    gsl_spline_init(splineD, zetaSorted, dSorted, nptsRaw);
    
    accLambda = gsl_interp_accel_alloc();
    splineLambda = gsl_spline_alloc(gsl_interp_cspline, nptsRaw);
    gsl_spline_init(splineLambda, zetaSorted, lambdaSorted, nptsRaw);
    
    accAlphax = gsl_interp_accel_alloc();
    splineAlphax = gsl_spline_alloc(gsl_interp_cspline, nptsRaw);
    gsl_spline_init(splineAlphax, zetaSorted, alphaxSorted, nptsRaw);
    
    accAlphaz = gsl_interp_accel_alloc();
    splineAlphaz = gsl_spline_alloc(gsl_interp_cspline, nptsRaw);
    gsl_spline_init(splineAlphaz, zetaSorted, alphazSorted, nptsRaw);
    
    accThetap = gsl_interp_accel_alloc();
    splineThetap = gsl_spline_alloc(gsl_interp_cspline, nptsRaw);
    gsl_spline_init(splineThetap, zetaSorted, thetapSorted, nptsRaw);
    
    // resample hypersurface variables on a uniform grid in zeta
    double dzeta = M_PI/2/(ZETAPTS-1);
    for (int i=0;i<ZETAPTS;i++) {
        zeta[i] = dzeta*i;
        d[i] = gsl_spline_eval(splineD, zeta[i], accD);
        L[i] = gsl_spline_eval(splineLambda, zeta[i], accLambda);
        AX[i] = gsl_spline_eval(splineAlphax, zeta[i], accAlphax);
        AZ[i] = gsl_spline_eval(splineAlphaz, zeta[i], accAlphaz);
        TP[i] = gsl_spline_eval(splineThetap, zeta[i], accThetap);
    }
    // load derivative of d
    for (int i=0;i<ZETAPTS;i++) dp[i]=deriv(zeta,d,i);
    
    // free interpolating functions
    gsl_spline_free(splineD);
    gsl_interp_accel_free(accD);
    
    gsl_spline_free(splineLambda);
    gsl_interp_accel_free(accLambda);
    
    gsl_spline_free(splineAlphax);
    gsl_interp_accel_free(accAlphax);
    
    gsl_spline_free(splineAlphaz);
    gsl_interp_accel_free(accAlphaz);
    
    gsl_spline_free(splineThetap);
    gsl_interp_accel_free(accThetap);
    
    delete zetaRaw;
    delete zetaSorted;
    delete dRaw;
    delete dSorted;
    delete lambdaSorted;
    delete alphaxSorted;
    delete alphazSorted;
    delete thetapSorted;
}

double computeFreezeoutIntegral(double m, double pt, double y, double phi, double stats) {
    
    double thetaMax = M_PI - THETAMIN;
    double dTheta = (thetaMax-THETAMIN)/(THETAPTS-1);
    double dPhi = 2*M_PI/(PHIPTS-1);
    double dZeta = M_PI/2/(ZETAPTS-1);
    
    particleData pd = {sqrt(m*m+pt*pt), pt, y, phi, stats};
    
    double res = 0;
    for (int i=0;i<THETAPTS;i++) {
        double theta = THETAMIN + i*dTheta;
        double dtheta = (i==0 || i==THETAPTS-1 ? dTheta/2 : dTheta); // trapezoidal rule
        for (int j=0;j<PHIPTS;j++) {
            double phi = j*dPhi;
            double dphi = (j==0 || j==PHIPTS-1 ? dPhi/2 : dPhi); // trapezoidal rule
            for (int k=0;k<ZETAPTS;k++) {
                double dzeta = (k==0 || k==ZETAPTS-1 ? dZeta/2 : dZeta); // trapezoidal rule
                fluidVars fv = {L[k], AX[k], AZ[k], TP[k]};
                hypersurfaceVars hs = {d[k], dp[k], zeta[k], phi, theta};
                res += dzeta*dphi*dtheta*freezeoutIntegrand(pd,fv,hs);
            }
        }
    }
    
    return res/pow(2*M_PI,3);
}

void doFreezeOut() {
    
    print_line();
    cout << "==> Performing freezeout" << endl;
    
    outputRawFreezeoutData();
    
    cout << "--> Constructing hypersurface data" << endl;
    loadFreezeoutVariables();
    outputSampledFreezeoutData();

    fstream particle_number_out;
    particle_number_out.open("output/particle_number.dat", ios::out);
    double chargedParticleNumber = 0;
    
    fstream spectra_out;
    
    ifstream inputFile;
    inputFile.open(PARTICLEFILE);
    string line;
    getline(inputFile, line); // read and discard the first line (header)
    while (getline(inputFile, line))
    {
        istringstream ss(line);
        char particleName[30];
        double massGeV;
        double charge;
        double spin;
        double statistics;
        ss >> particleName >> massGeV >> charge >> spin >> statistics;
        cout << "--> Processing " << particleName << "; ";
        
        double mass = massGeV/HBARC;
        
        char fname[100];
    
/*    
        sprintf(fname,"output/spectrum_%s_phi.dat",particleName);
        spectra_out.open(fname, ios::out);
        int phisteps = 20;
        double dphi = 2*M_PI/(phisteps-1);
        for (int i=0; i<phisteps; i++) {
            double phi = i*dphi;
            double res = computeFreezeoutIntegral(mass,1.,0.,phi);
            spectra_out << phi << "\t" << res << endl;
        }
        spectra_out.close();
        
        sprintf(fname,"output/spectrum_%s_y.dat",particleName);
        spectra_out.open(fname, ios::out);
        int ysteps = 20;
        double dy = 6./(ysteps-1);
        for (int i=0; i<ysteps; i++) {
            double y = -3 + i*dy;
            double res = computeFreezeoutIntegral(mass,1.,y,0.);
            spectra_out << y << "\t" << res << endl;
        }
        spectra_out.close();
*/
       
        sprintf(fname,"output/spectrum_%s_pt.dat",particleName);
        spectra_out.open(fname, ios::out);
        // outputs two columns: pt [GeV], dN/(2 Pi pt dpt dy)|_y=0 [1/GeV^2]
        int lptsteps = 40;
        double lptmin = -2;
        double lptmax = 1.5;
        double dlpt = (lptmax-lptmin)/(lptsteps-1);
        double temp = 0;
        for (int i=0; i<lptsteps; i++) {
            double pt = pow(10,lptmin + i*dlpt);
            double res = computeFreezeoutIntegral(mass,pt,0.,0.,statistics);
            temp += pt*pt*res;
            spectra_out << pt*HBARC << "\t" << res/HBARC/HBARC << endl;
        }
        temp *= 2*M_PI*(2*spin+1)*log(10.)*dlpt;
        spectra_out.close();
        
        cout << "dN/dy = " << temp << endl;
        if (charge!=0) chargedParticleNumber += temp;
    }
    inputFile.close();
    cout << "==> Total charged particle dN/dy =  " << chargedParticleNumber << endl;
    
    particle_number_out << ETABAR << "\t" << chargedParticleNumber << endl;
    particle_number_out.close();
}

//////////////////////////////////////////////////////////////////////////////////////
// output/debugging functions
//////////////////////////////////////////////////////////////////////////////////////

void outputSampledFreezeoutData() {
    
    fstream fout;
    fout.open("output/hypersurface_zeta.dat", ios::out);
    for (int i = 0; i < ZETAPTS; i++)
    {
        fout << zeta[i] << "\t";
        fout << d[i]    << "\t";
        fout << dp[i]   << "\t";
        fout << L[i]    << "\t";
        fout << AX[i]   << "\t";
        fout << AZ[i]   << "\t";
        fout << TP[i]   << "\t";
        fout << Teff(AX[i],AZ[i],L[i]) << endl;
    }
    fout.close();
    
}

void outputRawFreezeoutData() {
    
    fstream fout;
    fout.open("output/hypersurface.dat", ios::out);
    for (vector<double>::size_type i = 0; i < tauVec.size(); i++)
    {
        fout << rfoVec[i] << "\t" << tauVec[i] << endl;
    }
    fout.close();
    
    fout.open("output/hypersurface_variables.dat", ios::out);
    for (vector<double>::size_type i = 0; i < tauVec.size(); i++)
    {
        fout << tauVec[i] << "\t" << rfoVec[i] << "\t" << lambdaVec[i] << "\t" << alphaxVec[i] << "\t" << alphazVec[i] << "\t" << thetapVec[i] << endl;
    }
    fout.close();
    
}
