/*

   glauber.cpp

   Copyright (c) Mohammad Nopoush and Michael Strickland

   GNU General Public License (GPLv3)
*/

#include <iostream>
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include <cmath>
#include <gsl/gsl_integration.h>

using namespace std;

#include "glauber.h"
#include "aHydro1p1b.h"
#include "eos.h"

struct ta_params { double x; double y; double A;};

double woodsSaxonDist(double r, double A) {
	double n0 = 0.17; // central density in fm^(-3)
	double Rn = 1.12*pow(A,1./3.) - 0.86*pow(A,-1./3.); // radius in fm 
	double d = 0.54; // thickness in fm
	return n0/(1+exp((r-Rn)/d));
}

// generates A samples corresponding to Woods-Saxon nucleus with A nucleons
// results are loaded into x and y arrays (z is not used in current context)
// note: assumes that the the random number generator has already been seeded
void sampleWoodsSaxon(int A, double *x, double *y) {
	const double M = 1.01*4.5310551374155095;
	const double rmax = 20;
	int m = 0;
	double r,v,f,u;
	while (m<A) {
	  r = rmax*((double) rand())/((double)RAND_MAX);
	  u = ((double) rand())/((double)RAND_MAX);
	  if (u < r*r*woodsSaxonDist(r,A)/M) {
	    v = 2*(((double) rand())/((double)RAND_MAX)-0.5);
	    f = 2*M_PI*((double) rand())/((double)RAND_MAX);
	    x[m] = r*sqrt(1-v*v)*cos(f);
	    y[m] = r*sqrt(1-v*v)*sin(f);
	    m++;
	  }
	}
}

// finds colliding nucleons using hard sphere model
// returns number of wounded nucleons found and loads their
// positions into the x and y arrays passed down to it
int MCcollisions(int A, double b, double *x, double *y) {
	double *x1,*y1,*x2,*y2,dn,dist;	
	int *l1,*l2;
	x1 = new double[A];
	y1 = new double[A];
	l1 = new int[A];
	x2 = new double[A];
	y2 = new double[A];
	l2 = new int[A];
	sampleWoodsSaxon(A,x1,y1);
	sampleWoodsSaxon(A,x2,y2);
	for (int i=0; i<A; i++) {
		x1[i] -= b/2;
		x2[i] += b/2;
	}
	dn = sqrt(0.1*SIGMANN/M_PI);
	for (int i=0; i<A; i++) {
		l1[i] = 0;
		l2[i] = 0;
	}
	for (int i=0; i<A; i++)
	  for (int j=0; j<A; j++) {
	    dist = pow(x1[i]-x2[j],2);
	    dist += pow(y1[i]-y2[j],2);
	    dist = sqrt(dist);
	    if (dist<dn) {
		l1[i] += 1;
		l2[j] += 1;
	    }
	  }
	int n = 0;
	for (int i=0; i<A; i++) {
	  if (l1[i]>0) {
	    x[n] = x1[i];
	    y[n] = y1[i];
	    n++;
	  }
	  if (l2[i]>0) {
	    x[n] = x2[i];
	    y[n] = y2[i];
	    n++;
	  }
	}
	delete[] x1;
	delete[] y1;
	delete[] l1;
	delete[] x2;
	delete[] y2;
	delete[] l2;
	return n;
}

double TAintegrand(double z, void * params) {
	struct ta_params * my_params = (struct ta_params *)params;
        double x = my_params->x;
        double y = my_params->y;
        double A = my_params->A;
	return woodsSaxonDist(sqrt(x*x+y*y+z*z),A);
}


// thickness function
double TA(double x, double y, double A) {
	double result, error;
	double n = 100;
        struct ta_params int_params = {x,y,A};
	gsl_function F;
        F.function = &TAintegrand;
        F.params = &int_params;
	gsl_integration_workspace * w = gsl_integration_workspace_alloc(n);
	gsl_integration_qags(&F, 0., 50., 0, 1.0e-6, n, w, &result, &error);
	gsl_integration_workspace_free(w);
	return 2*result;
}

double TAB(double x, double y, double A, double B, double b) {
	return TA(x+b/2,y,A)*TA(x-b/2,y,B);
}

double nPartAB(double x, double y, double A, double B, double b) {
	// 1 fm^(-2) = 0.1 mb^(-1)
	double t,rval;
	t = SIGMANN*TA(x-b/2,y,B)*0.1/B;
	if (t > SMALLT)
		rval = TA(x+b/2,y,A)*(1. - pow(1. - t,(double)B));
	else
		rval = TA(x+b/2,y,A)*(B*t*(6 + (-1 + B)*t*(-3 + (-2 + B)*t)))/6.;
	t = SIGMANN*TA(x+b/2,y,A)*0.1/A;
	if (t > SMALLT)
		rval += TA(x-b/2,y,B)*(1. - pow(1. - t,(double)A));
	else
		rval += TA(x-b/2,y,B)*(A*t*(6 + (-1 + A)*t*(-3 + (-2 + A)*t)))/6.;
	return rval;
}

// wounded nucleon energy density profile
double edProfileABpart(double x, double y, double A, double B, double b) {
        return nPartAB(x,y,A,B,b)/nPartAB(0.,0.,A,B,0.);
}

// binary collision energy density profile
double edProfileABbin(double x, double y, double A, double B, double b) {
        return TAB(x,y,A,B,b)/TAB(0.,0.,A,B,0.);
}
