/*
 
 gubser.cpp
 
 Copyright (c) Mohammad Nopoush and Michael Strickland
 
 GNU General Public License (GPLv3)
 See detailed text in license directory
 
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_odeiv2.h>
#include <gsl/gsl_spline.h>

using namespace std;

#include "aHydro1p1b.h"
#include "rfuncs.h"
#include "gubser.h"

#define Power(a,b) pow((double)a,(double)b)

double rhoPts[NRHOSTEPS+1];
double LSol[NRHOSTEPS+1];
double azSol[NRHOSTEPS+1];

gsl_interp_accel *accL;
gsl_spline *splineL;

gsl_interp_accel *accAZ;
gsl_spline *splineAZ;

int aHydroGubserEQs(double rho, const double y[], double f[], void *params)
{
    double L = y[0];
    double az = y[1];
    
    double T = pow(Rb(az),0.25)*L;
    double teq = 5*ETABAR/T;
    double yb = sqrt((3*az*az-1)/2);
    double h2 = H2(yb);
    double h2T = H2T(yb);
    double h2L = H2L(yb);
    double tanhrho = tanh(rho);
    
    f[0] = -((2*h2 + h2T)*L*tanhrho)/(4.*h2) - ((-4 + (3*Power(az,2)*(h2 + h2L))/h2)*(-3*(1 - 4*Power(az,2) + 3*Power(az,4))*Power(T,5) + 8*Power(az,5)*Power(L,5)*tanhrho*teq))/(96.*Power(az,7)*Power(L,4)*teq);
    f[1] = ((-1 + 3*Power(az,2))*(8*Power(az,5)*tanhrho - (3*(1 - 4*Power(az,2) + 3*Power(az,4))*Power(T,5))/(Power(L,5)*teq)))/(24.*Power(az,6));
    
    return GSL_SUCCESS;
}

void solveGubserEqs(double lambda0, double alphaz0) {
    
    gsl_odeiv2_system sys = {aHydroGubserEQs, NULL, 2, NULL};
    
    gsl_odeiv2_driver *d = gsl_odeiv2_driver_alloc_y_new(&sys, gsl_odeiv2_step_rk8pd, 1e-12, 0, 1e-12);
    
    int i;
    double rhoi = -RHOMAX;
    double drho = 2*RHOMAX/((double) NRHOSTEPS);
    double y[2] = { lambda0, alphaz0 };
    
    rhoPts[0] = rhoi;
    LSol[0] = lambda0;
    azSol[0] = alphaz0;
    
    for (i = 1; i <= NRHOSTEPS; i++)
    {
        double rho = -RHOMAX + drho*i;
        int status = gsl_odeiv2_driver_apply(d, &rhoi, rho, y);
        if (status != GSL_SUCCESS)
        {
            cout << "error, return value=" << status << endl;
            break;
        }
        rhoPts[i] = rho;
        LSol[i] = y[0];
        azSol[i] = y[1];
    }
    
    gsl_odeiv2_driver_free(d);
    
}

void setupGubserInterpolations() {
    // setup lambda interpolation
    accL = gsl_interp_accel_alloc();
    splineL = gsl_spline_alloc(gsl_interp_cspline, NRHOSTEPS+1);
    gsl_spline_init(splineL, rhoPts, LSol, NRHOSTEPS+1);
    // setup alphaz interpolation
    accAZ = gsl_interp_accel_alloc();
    splineAZ = gsl_spline_alloc(gsl_interp_cspline, NRHOSTEPS+1);
    gsl_spline_init(splineAZ, rhoPts, azSol, NRHOSTEPS+1);
}

inline double LSolInt(double rho) {
    return gsl_spline_eval(splineL, rho, accL);
}

inline double azSolInt(double rho) {
    return gsl_spline_eval(splineAZ, rho, accAZ);
}

void freeGubserInterpolations() {
    // free lambda interpolation
    gsl_spline_free(splineL);
    gsl_interp_accel_free(accL);
    // free alphaz interpolation
    gsl_spline_free(splineAZ);
    gsl_interp_accel_free(accAZ);
}

double lambdaGubserMinkowski(double t, double r) {
    return LSolInt(-asinh((1 - Q*Q*t*t + Q*Q*r*r)/2/Q/t))/t;
}

double azGubserMinkowski(double t, double r) {
    return azSolInt(-asinh((1 - Q*Q*t*t + Q*Q*r*r)/2/Q/t));
}

double thetapGubserMinkowski(double t, double r) {
    return atanh(2*Q*Q*r*t/(1+Q*Q*t*t+Q*Q*r*r));
}
