/*
 
 initialconditions.cpp
 
 Copyright (c) Mohammad Nopoush and Michael Strickland
 
 GNU General Public License (GPLv3)
 See detailed text in license directory
 
 */

#include <iostream>
#include <fstream>
#include <cmath>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_odeiv2.h>
#include <gsl/gsl_spline.h>

using namespace std;

#include "aHydro1p1b.h"
#include "initialconditions.h"
#include "gubser.h"
#include "glauber.h"
#include "eos.h"
#include "rfuncs.h"

// initializes the variables
void setInitialConditions()
{
    
    cout << "==> Initializing variables\n";
  
    double r; // radius

    //double ALPHAX0 = sqrt(2.)*ALPHAZ0/sqrt(3.*ALPHAZ0*ALPHAZ0-1.);
    double centralR = R(ALPHAX0,ALPHAZ0);
    double centralED = Eiso(L0)*centralR;

    fstream checkFile; // used for saving exact Gubser solution

    double small = 1.e-20; // used below as energy density offset at large r 
    
    switch (INITCONDTYPE) {
        case 0: {
            // transverally homogeneous (constant)
            cout << "==> Initial condition : transversally homogeneous" << endl;
            for (int s=0;s<NUM;s++) {
                r = (s - (NUM-1)/2.)*A;
                lambda[s] = L0;
                alphaz[s] = ALPHAZ0;
                alphax[s] = ALPHAX0;
                thetap[s] = r*VERYSMALL;
            }
            return;
        }
        case 1: {
            // aHydro Gubser
            cout << "==> Initial condition : Gubser flow exact aHydro solution" << endl;
            solveGubserEqs(L0,ALPHAZ0); // solve the aHydro-Gubser ODEs
            setupGubserInterpolations(); // construct interpolating functions
            for (int s=0;s<NUM;s++) {
                r = (s - (NUM-1)/2.)*A;
                lambda[s] = lambdaGubserMinkowski(TAU0,r);
                alphaz[s] = azGubserMinkowski(TAU0,r);
                alphax[s] = sqrt(2.)*alphaz[s]/sqrt(3.*alphaz[s]*alphaz[s]-1.);
                thetap[s] = thetapGubserMinkowski(TAU0,r);
            }
            //
            // output results at final time for later comparison
            //
            double tauf = TAU0 + STEPS*EPS;
            // lambda
            checkFile.open("output/LSol.dat", ios::out);
            for (int s=0;s<NUM;s++) {
                r = (s - ((double)NUM-1)/2.)*A;
                checkFile << r << "\t";
                checkFile << lambdaGubserMinkowski(tauf,r) << endl;
            }
            checkFile.close();
            // az
            checkFile.open("output/azSol.dat", ios::out);
            for (int s=0;s<NUM;s++) {
                r = (s - ((double)NUM-1)/2.)*A;
                checkFile << r << "\t";
                checkFile << azGubserMinkowski(tauf,r) << endl;
            }
            checkFile.close();
            // thetap
            checkFile.open("output/thetapSol.dat", ios::out);
            for (int s=0;s<NUM;s++) {
                r = (s - ((double)NUM-1)/2.)*A;
                checkFile << r << "\t";
                checkFile << thetapGubserMinkowski(tauf,r) << endl;
            }
            checkFile.close();
            freeGubserInterpolations(); // free the interpolating functions
            return;
        }
        case 2: {
            // glauber participant with impact parameter 0
            cout << "==> Initial condition : Glauber wounded nucleon, isotropic, no flow" << endl;
            for (int s=0;s<NUM;s++) {
                r = (s - (NUM-1)/2.)*A;
                lambda[s] = Teff(((centralED-small)*edProfileABpart(r,0,(double)NA,(double)NB,0.) + small)/centralR);
                alphaz[s] = ALPHAZ0;
                alphax[s] = ALPHAX0;
                thetap[s] = r*VERYSMALL;
            }
            break;
        }
        case 3: {
            // glauber binary with impact parameter 0
            cout << "==> Initial condition : Glauber wounded nucleon, isotropic, no flow" << endl;
            for (int s=0;s<NUM;s++) {
                r = (s - (NUM-1)/2.)*A;
                lambda[s] = Teff(((centralED-small)*edProfileABbin(r,0,(double)NA,(double)NB,0.) + small)/centralR);
                alphaz[s] = ALPHAZ0;
                alphax[s] = ALPHAX0;
                thetap[s] = r*VERYSMALL;
            }
            break;
        }
        default: {
            cout << "Invalid initial condition type" << endl;
            exit(0);
            break;
        }
    }
    
}
