/*
 
 rfuncs.cpp
 
 Copyright (c) Mohammed Nopoush and Michael Strickland
 
 GNU General Public License (GPLv3)
 See detailed text in license directory
 
 */

#include <iostream>
#include <iomanip>
#include <cmath>
#include <gsl/gsl_integration.h>

#include "rfuncs.h"

using namespace std;

#define Power(a,b) pow((double)a,(double)b)
#define ArcTan(a) atan((double)a)
#define ArcTanh(a) atanh((double)a)
#define Sqrt(a) sqrt((double)a)
#define Abs(a) fabs((double)a)

// external params for integration

struct integralParams {
    double ax, az;
};

struct functionWrapper {
    double (* function) (double x, void * params);
};

// special functions

double ay(double ax, double az) {
    return ax*az/sqrt(3*pow(ax*az,2) - az*az - ax*ax);
}

double d1ay(double ax, double az) {
    return -(Power(az,3)/Power(-Power(az,2) + Power(ax,2)*(-1 + 3*Power(az,2)),1.5));
}

double d2ay(double ax, double az) {
    return -(Power(ax,3)/Power(-Power(az,2) + Power(ax,2)*(-1 + 3*Power(az,2)),1.5));
}

inline double ap(double ax, double az, double phi) {
    return sqrt(pow(ax*cos(phi),2)+pow(ay(ax,az)*sin(phi),2));
}

//************************************************************************
// H functions
//************************************************************************

double H2(double u) {
    if (u==0) return 0;
    if (u==1) return 2;
    if (fabs(u-1)<SMALL)
        return 2 + (8*(-1 + u))/3. + (4*Power(-1 + u,2))/5. + (8*Power(-1 + u,3))/105. - (2*Power(-1 + u,4))/63. + (16*Power(-1 + u,5))/1155. - (8*Power(-1 + u,6))/1287.;
    if (abs(u)<1)
        return (u*(u*Sqrt(1 - Power(u,2)) + ArcTan(Sqrt(1 - Power(u,2))/u)))/Sqrt(1 - Power(u,2));
    else
        return (u*(u*Sqrt(-1 + Power(u,2)) + ArcTanh(Sqrt(-1 + Power(u,2))/u)))/Sqrt(-1 + Power(u,2));
}

double H2T(double u) {
    if (u==0) return 0;
    if (u==1) return 4./3.;
    if (fabs(u-1)<SMALL)
        return 4./3. + (16*(-1 + u))/15. - (8*Power(-1 + u,2))/35. + (16*Power(-1 + u,3))/315. - (4*Power(-1 + u,4))/693. - (64*Power(-1 + u,5))/15015. + (32*Power(-1 + u,6))/6435.;
    if (abs(u)<1)
        return -((u*(-(u*Sqrt(1 - Power(u,2))) + (-1 + 2*Power(u,2))*ArcTan(Sqrt(1 - Power(u,2))/u)))/Power(1 - Power(u,2),1.5));
    else
        return (u*(-(u*Sqrt(-1 + Power(u,2))) + (-1 + 2*Power(u,2))*ArcTanh(Sqrt(-1 + Power(u,2))/u)))/Power(-1 + Power(u,2),1.5);
}

double H2L(double u) {
    if (u==0) return 0;
    if (u==1) return 2./3.;
    if (fabs(u-1)<SMALL)
        return 2./3. + (8*(-1 + u))/5. + (36*Power(-1 + u,2))/35. + (8*Power(-1 + u,3))/315. - (2*Power(-1 + u,4))/77. + (272*Power(-1 + u,5))/15015. - (8*Power(-1 + u,6))/715.;
    if (abs(u)<1)
        return -((Power(u,3)*(u*Sqrt(1 - Power(u,2)) - ArcTan(Sqrt(1 - Power(u,2))/u)))/Power(1 - Power(u,2),1.5));
    else
        return (Power(u,3)*(u*Sqrt(-1 + Power(u,2)) - ArcTanh(Sqrt(-1 + Power(u,2))/u)))/Power(-1 + Power(u,2),1.5);
}

//************************************************************************
// Integrands
//************************************************************************

double R_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ap(ax,az,phi),2)*H2(az/ap(ax,az,phi));
}

double Rx_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ax*cos(phi),2)*H2(az/ap(ax,az,phi));
}

double Ry_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ay(ax,az)*sin(phi),2)*H2(az/ap(ax,az,phi));
}

double HTx_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ax*cos(phi),2)*H2T(az/ap(ax,az,phi));
}

double HTy_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ay(ax,az)*sin(phi),2)*H2T(az/ap(ax,az,phi));
}

double HL_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ap(ax,az,phi),2)*H2L(az/ap(ax,az,phi));
}


double HLx_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ax*cos(phi),2)*H2L(az/ap(ax,az,phi));
}

double HLy_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ay(ax,az)*sin(phi),2)*H2L(az/ap(ax,az,phi));
}

double d1HTx_s_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ax*cos(phi),2)/pow(ap(ax,az,phi),2)*(ax*cos(phi)*cos(phi)+ay(ax,az)*d1ay(ax,az)*sin(phi)*sin(phi))*(1.0666666666666667 + (64*(-1 + az/ap(ax,az,phi)))/105. - (32*Power(-1 + az/ap(ax,az,phi),2))/105. + (64*Power(-1 + az/ap(ax,az,phi),3))/495. - (400*Power(-1 + az/ap(ax,az,phi),4))/9009. +
                                                                                                                   (128*Power(-1 + az/ap(ax,az,phi),5))/15015.);
}

double d1HTx_l_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ax*cos(phi),2)/(az*az-pow(ap(ax,az,phi),2))*(2*H2L(az/ap(ax,az,phi))-H2T(az/ap(ax,az,phi)))*(ax*cos(phi)*cos(phi)+ay(ax,az)*d1ay(ax,az)*sin(phi)*sin(phi));
}


double d2HTx_s_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ax*cos(phi),2)/az*(1-az*ay(ax,az)/pow(ap(ax,az,phi),2)*d2ay(ax,az)*sin(phi)*sin(phi))*(1.0666666666666667 + (64*(-1 + az/ap(ax,az,phi)))/105. - (32*Power(-1 + az/ap(ax,az,phi),2))/105. + (64*Power(-1 + az/ap(ax,az,phi),3))/495. - (400*Power(-1 + az/ap(ax,az,phi),4))/9009. +
                                                                                                      (128*Power(-1 + az/ap(ax,az,phi),5))/15015.);
}

double d2HTx_l_integrand(double phi, void *params) {
    
    struct integralParams *my_params = (struct integralParams *) params;
    double ax = my_params->ax;
    double az = my_params->az;
    
    return pow(ax*ap(ax,az,phi)*cos(phi),2)/az/(az*az-pow(ap(ax,az,phi),2))*(2*H2L(az/ap(ax,az,phi))-H2T(az/ap(ax,az,phi)))*(1-az*ay(ax,az)/pow(ap(ax,az,phi),2)*d2ay(ax,az)*sin(phi)*sin(phi));
}

double integrate(functionWrapper * fw, double ax, double az) {
    
    gsl_integration_workspace *w = gsl_integration_workspace_alloc(NPTS);
    double result, error;
    struct integralParams params = { ax, az };
    
    gsl_function F;
    F.function = fw->function;
    F.params = &params;
    
    gsl_integration_qag(&F, 0, 0.5*M_PI, 0, TOLERANCE, NPTS, GSL_INTEG_GAUSS21, w, &result, &error);
    
    gsl_integration_workspace_free(w);
    
    return result*ax*ay(ax,az)/(M_PI);
}

double integrate_s(functionWrapper * fw_s, double ax, double az) {
    
    gsl_integration_workspace *w = gsl_integration_workspace_alloc(NPTS);
    double result, error;
    struct integralParams params = { ax, az };
    
    gsl_function F;
    F.function = fw_s->function;
    F.params = &params;
    
    gsl_integration_qag(&F, 0, 0.01, 0, TOLERANCE, NPTS, GSL_INTEG_GAUSS21, w, &result, &error);
    
    gsl_integration_workspace_free(w);
    
    return result*ax*ay(ax,az)/(M_PI);
}

double integrate_l(functionWrapper * fw_l, double ax, double az) {
    
    gsl_integration_workspace *w = gsl_integration_workspace_alloc(NPTS);
    double result, error;
    struct integralParams params = { ax, az };
    
    gsl_function F;
    F.function = fw_l->function;
    F.params = &params;
    
    gsl_integration_qag(&F, 0.01, 0.5*M_PI, 0, TOLERANCE, NPTS, GSL_INTEG_GAUSS21, w, &result, &error);
    
    gsl_integration_workspace_free(w);
    
    return result*ax*ay(ax,az)/(M_PI);
}

//************************************************************************
// R, Rx, Ry, dR/dax, dR/daz, HTx, HTy, dHT/dax, dHT/adz, HL, HLx, HLy
//************************************************************************

double R(double ax, double az) {
    if (ax==1 && az==1) return 1.;
    functionWrapper fw;
    fw.function = &R_integrand;
    return integrate(&fw,ax,az);
}


double Rx(double ax, double az) {
    if (ax==1 && az==1) return 0.5;
    functionWrapper fw;
    fw.function = &Rx_integrand;
    return integrate(&fw,ax,az);
}

double Ry(double ax, double az) {
    if (ax==1 && az==1) return 0.5;
    functionWrapper fw;
    fw.function = &Ry_integrand;
    return integrate(&fw,ax,az);
}

double d1R(double ax, double az) {
    if (ax==1 && az==1) return 0.;
    functionWrapper fw;
    return (1/ax+1/ay(ax,az)*d1ay(ax,az))*R(ax,az)+HTx(ax,az)/ax+HTy(ax,az)/ay(ax,az)*d1ay(ax,az);
}

double d2R(double ax, double az) {
    if (ax==1 && az==1) return 0.;
    functionWrapper fw;
    return 1/ay(ax,az)*d2ay(ax,az)*(R(ax,az)+HTy(ax,az))+1/az*(R(ax,az)+HL(ax,az));
}

double HTx(double ax, double az) {
    if (ax==1 && az==1) return 1./3.;
    functionWrapper fw;
    fw.function = &HTx_integrand;
    return integrate(&fw,ax,az);
}

double HTy(double ax, double az) {
    if (ax==1 && az==1) return 1./3.;
    functionWrapper fw;
    fw.function = &HTy_integrand;
    return integrate(&fw,ax,az);
}


double d1HTx(double ax, double az) {
    if (ax==1 && az==1) return 8./15.;
    functionWrapper fw_s;
    functionWrapper fw_l;
    fw_s.function = &d1HTx_s_integrand;
    fw_l.function = &d1HTx_l_integrand;
    return (3/ax+1/ay(ax,az)*d1ay(ax,az))*HTx(ax,az)-integrate_s(&fw_s,ax,az)-integrate_l(&fw_l,ax,az);
}

double d2HTx(double ax, double az) {
    if (ax==1 && az==1) return 0.;
    functionWrapper fw_s;
    functionWrapper fw_l;
    fw_s.function = &d2HTx_s_integrand;
    fw_l.function = &d2HTx_l_integrand;
    return 1/ay(ax,az)*d2ay(ax,az)*HTx(ax,az)+integrate_s(&fw_s,ax,az)+integrate_l(&fw_l,ax,az);
}

double HL(double ax, double az) {
    if (ax==1 && az==1) return 1./3.;
    functionWrapper fw;
    fw.function = &HL_integrand;
    return integrate(&fw,ax,az);
}

double HLx(double ax, double az) {
    if (ax==1 && az==1) return 1./6.;
    functionWrapper fw;
    fw.function = &HLx_integrand;
    return integrate(&fw,ax,az);
}

double HLy(double ax, double az) {
    if (ax==1 && az==1) return 1./6.;
    functionWrapper fw;
    fw.function = &HLy_integrand;
    return integrate(&fw,ax,az);
}

double Rb(double az) {
    double yb = sqrt((3*az*az-1)/2.);
    return pow(az/yb,4.)*H2(yb)/2.;
}


